import {
  useFreshRef
} from "./chunk-KTZM67PO.js";
import {
  getErrorMessageFromFramesStackItem,
  isPartialFrameStackItem
} from "./chunk-4MNUBRTU.js";

// src/ui/frame.base.tsx
import {
  createElement as reactCreateElement,
  useCallback,
  useEffect,
  useRef,
  useState
} from "react";
function defaultMessageHandler() {
}
function defaultErrorLogger(error) {
  console.error(error);
}
function BaseFrameUI({
  frameState,
  components,
  theme,
  allowPartialFrame = false,
  enableImageDebugging = false,
  onError = defaultErrorLogger,
  onMessage = defaultMessageHandler,
  createElement = reactCreateElement
}) {
  const [isImageLoading, setIsImageLoading] = useState(true);
  const { currentFrameStackItem } = frameState;
  const rootRef = useRef(null);
  const rootDimensionsRef = useRef();
  const onErrorRef = useFreshRef(onError);
  const onImageLoadEnd = useCallback(() => {
    setIsImageLoading(false);
  }, []);
  useEffect(() => {
    if (currentFrameStackItem?.status === "pending") {
      setIsImageLoading(true);
    }
  }, [currentFrameStackItem?.status]);
  useEffect(() => {
    if (currentFrameStackItem?.status === "requestError") {
      onErrorRef.current(currentFrameStackItem.requestError);
    }
  }, [currentFrameStackItem, onErrorRef]);
  const onMessageRef = useFreshRef(onMessage);
  useEffect(() => {
    if (currentFrameStackItem?.status === "message") {
      onMessageRef.current({
        message: currentFrameStackItem.message,
        status: currentFrameStackItem.type === "info" ? "message" : "error"
      });
    }
  }, [currentFrameStackItem, onMessageRef]);
  if (!frameState.homeframeUrl) {
    return components.Error(
      { message: "Missing frame url" },
      theme?.Error || {}
    );
  }
  if (!currentFrameStackItem) {
    return null;
  }
  let frameUiState;
  const previousFrameStackItem = frameState.framesStack[frameState.framesStack.length - 1];
  const previousFrame = previousFrameStackItem?.status === "done" && previousFrameStackItem.frameResult.specification !== "farcaster_v2" ? previousFrameStackItem.frameResult.frame : null;
  switch (currentFrameStackItem.status) {
    case "requestError": {
      if ("sourceFrame" in currentFrameStackItem.request && currentFrameStackItem.request.sourceFrame) {
        frameUiState = {
          status: "complete",
          frame: currentFrameStackItem.request.sourceFrame,
          isImageLoading,
          id: currentFrameStackItem.id,
          frameState
        };
      } else {
        return components.Error(
          {
            message: "Failed to load frame",
            error: currentFrameStackItem.requestError
          },
          theme?.Error || {}
        );
      }
      break;
    }
    case "message":
      if (!currentFrameStackItem.request.sourceFrame) {
        return components.Message(
          {
            message: currentFrameStackItem.message,
            status: currentFrameStackItem.type === "error" ? "error" : "message"
          },
          theme?.Message ?? {}
        );
      }
      frameUiState = {
        status: "complete",
        frame: currentFrameStackItem.request.sourceFrame,
        isImageLoading,
        id: currentFrameStackItem.id,
        frameState
      };
      break;
    case "doneRedirect": {
      if (!currentFrameStackItem.request.sourceFrame) {
        frameUiState = {
          status: "loading",
          id: currentFrameStackItem.id,
          frameState
        };
      } else {
        frameUiState = {
          status: "complete",
          frame: currentFrameStackItem.request.sourceFrame,
          isImageLoading,
          id: currentFrameStackItem.id,
          frameState
        };
      }
      break;
    }
    case "done": {
      if (currentFrameStackItem.frameResult.status === "success") {
        if (currentFrameStackItem.frameResult.specification === "farcaster_v2") {
          frameUiState = {
            status: "complete",
            frame: currentFrameStackItem.frameResult.frame,
            specification: "farcaster_v2",
            debugImage: enableImageDebugging ? currentFrameStackItem.frameResult.framesDebugInfo?.image : void 0,
            isImageLoading,
            id: currentFrameStackItem.id,
            frameState,
            parseResult: currentFrameStackItem.frameResult
          };
        } else {
          frameUiState = {
            status: "complete",
            frame: currentFrameStackItem.frameResult.frame,
            debugImage: enableImageDebugging ? currentFrameStackItem.frameResult.framesDebugInfo?.image : void 0,
            isImageLoading,
            id: currentFrameStackItem.id,
            frameState
          };
        }
      } else if (isPartialFrameStackItem(currentFrameStackItem) && allowPartialFrame) {
        if (currentFrameStackItem.frameResult.specification === "farcaster_v2") {
          frameUiState = {
            status: "partial",
            frame: currentFrameStackItem.frameResult.frame,
            specification: "farcaster_v2",
            debugImage: enableImageDebugging ? currentFrameStackItem.frameResult.framesDebugInfo?.image : void 0,
            isImageLoading,
            id: currentFrameStackItem.id,
            frameState,
            parseResult: currentFrameStackItem.frameResult
          };
        } else {
          frameUiState = {
            status: "partial",
            frame: currentFrameStackItem.frameResult.frame,
            debugImage: enableImageDebugging ? currentFrameStackItem.frameResult.framesDebugInfo?.image : void 0,
            isImageLoading,
            id: currentFrameStackItem.id,
            frameState
          };
        }
      } else {
        return components.Error(
          { message: "Invalid frame" },
          theme?.Error || {}
        );
      }
      break;
    }
    case "pending": {
      frameUiState = {
        status: "loading",
        id: currentFrameStackItem.id,
        frameState
      };
      break;
    }
  }
  const isLoading = frameUiState.status === "loading" || frameUiState.isImageLoading;
  let buttonsProps = null;
  if (frameUiState.status !== "loading") {
    if ("specification" in frameUiState) {
      buttonsProps = [
        {
          frameState: frameUiState,
          frameButton: {
            action: "launch_frame",
            label: frameUiState.frame.button.title
          },
          index: 0,
          isDisabled: false,
          onPress() {
            try {
              if (!("onLaunchFrameButtonPress" in frameState)) {
                throw new Error(
                  "onLaunchFrameButtonPress is not implemented, you are porbably using old useFrame hook"
                );
              }
              frameState.onLaunchFrameButtonPress(
                frameUiState.status === "complete" ? {
                  status: "complete",
                  frame: frameUiState.frame,
                  parseResult: frameUiState.parseResult
                } : {
                  status: "partial",
                  frame: frameUiState.frame,
                  parseResult: frameUiState.parseResult
                }
              );
            } catch (e) {
              onErrorRef.current(e instanceof Error ? e : new Error(String(e)));
            }
          }
        }
      ];
    } else if (frameUiState.frame.buttons && frameUiState.frame.buttons.length > 0) {
      buttonsProps = frameUiState.frame.buttons.map((frameButton, index) => ({
        frameState: frameUiState,
        frameButton,
        index,
        isDisabled: false,
        onPress() {
          rootDimensionsRef.current = rootRef.current?.computeDimensions();
          Promise.resolve(
            frameState.onButtonPress(
              // @todo change the type onButtonPress to accept partial frame as well because that can happen if partial frames are enabled
              frameUiState.frame,
              frameButton,
              index
            )
          ).catch((e) => {
            onErrorRef.current(e instanceof Error ? e : new Error(String(e)));
          });
        }
      }));
    }
  }
  return components.Root(
    {
      createElement,
      frameState: frameUiState,
      dimensions: isLoading ? rootDimensionsRef.current ?? null : null,
      ref: rootRef,
      loadingScreen: isLoading ? components.LoadingScreen(
        {
          frameState: frameUiState,
          dimensions: rootDimensionsRef.current ?? null
        },
        theme?.LoadingScreen || {}
      ) : null,
      buttonsContainer: buttonsProps ? components.ButtonsContainer(
        {
          frameState: frameUiState,
          buttons: buttonsProps.map(
            (buttonProps) => components.Button(
              buttonProps,
              theme?.Button || {}
            )
          ),
          buttonsProps
        },
        theme?.ButtonsContainer || {}
      ) : null,
      imageContainer: components.ImageContainer(
        {
          frameState: frameUiState,
          aspectRatio: (
            // eslint-disable-next-line no-nested-ternary -- unnecessary to extract this to a variable
            (frameUiState.status === "loading" ? previousFrame?.imageAspectRatio : "specification" in frameUiState ? "3:2" : frameUiState.frame.imageAspectRatio) ?? "1.91:1"
          ),
          image: components.Image(
            frameUiState.status === "loading" ? {
              status: "frame-loading",
              frameState: frameUiState,
              onImageLoadEnd,
              aspectRatio: previousFrame?.imageAspectRatio ?? ("specification" in frameUiState ? "3:2" : void 0) ?? "1.91:1"
            } : {
              status: "frame-loading-complete",
              frameState: frameUiState,
              src: frameUiState.debugImage ?? ("specification" in frameUiState ? frameUiState.frame.imageUrl : frameUiState.frame.image),
              aspectRatio: ("specification" in frameUiState ? "3:2" : frameUiState.frame.imageAspectRatio) ?? "1.91:1",
              onImageLoadEnd
            },
            theme?.Image || {}
          ),
          messageTooltip: currentFrameStackItem.status === "message" ? components.MessageTooltip(
            {
              frameState: frameUiState,
              status: currentFrameStackItem.type === "error" ? "error" : "message",
              message: getErrorMessageFromFramesStackItem(
                currentFrameStackItem
              )
            },
            theme?.MessageTooltip || {}
          ) : null
        },
        theme?.ImageContainer || {}
      ),
      textInputContainer: frameUiState.status === "loading" || "specification" in frameUiState || typeof frameUiState.frame.inputText !== "string" ? null : components.TextInputContainer(
        {
          frame: frameUiState,
          textInput: components.TextInput(
            {
              // @TODO provide previous frame to pending state so we can remove loading check and render skeletons?
              isDisabled: false,
              frameState: frameUiState,
              placeholder: frameUiState.frame.inputText,
              onChange(text) {
                frameState.setInputText(text);
              },
              value: frameState.inputText
            },
            theme?.TextInput || {}
          )
        },
        theme?.TextInputContainer || {}
      )
    },
    theme?.Root || {}
  );
}

export {
  BaseFrameUI
};
