import {
  isSpecificationValid
} from "./chunk-F5WB6H3H.js";
import {
  tryCallAsync
} from "./chunk-RVPSX5WL.js";

// src/next/POST.tsx
import {
  getFrame
} from "frames.js";
import { parseFramesWithReports } from "frames.js/parseFramesWithReports";
import { z } from "zod";
var castActionMessageParser = z.object({
  type: z.literal("message"),
  message: z.string().min(1)
});
var castActionFrameParser = z.object({
  type: z.literal("frame"),
  frameUrl: z.string().min(1).url()
});
var composerActionFormParser = z.object({
  type: z.literal("form"),
  url: z.string().min(1).url(),
  title: z.string().min(1)
});
var jsonResponseParser = z.preprocess(
  (data) => {
    if (typeof data === "object" && data !== null && !("type" in data)) {
      return {
        type: "message",
        ...data
      };
    }
    return data;
  },
  z.discriminatedUnion("type", [
    castActionFrameParser,
    castActionMessageParser,
    composerActionFormParser
  ])
);
var errorResponseParser = z.object({
  message: z.string().min(1)
});
async function POST(req) {
  try {
    const searchParams = "nextUrl" in req ? req.nextUrl.searchParams : new URL(req.url).searchParams;
    const body = await req.json();
    const isPostRedirect = searchParams.get("postType") === "post_redirect";
    const isTransactionRequest = searchParams.get("postType") === "tx";
    const postUrl = searchParams.get("postUrl");
    const multiSpecificationEnabled = searchParams.get("multispecification") === "true";
    const specification = searchParams.get("specification") ?? "farcaster";
    if (!postUrl) {
      return Response.json(
        { message: "postUrl parameter not found" },
        {
          status: 400
        }
      );
    }
    const response = await fetch(postUrl, {
      method: "POST",
      headers: {
        Accept: "application/json",
        "Content-Type": "application/json"
      },
      redirect: isPostRedirect ? "manual" : void 0,
      body: JSON.stringify(body)
    });
    if (response.status >= 500) {
      const jsonError = await tryCallAsync(
        () => response.clone().json()
      );
      if (jsonError instanceof Error) {
        return Response.json(
          { message: jsonError.message },
          { status: response.status }
        );
      }
      const result2 = errorResponseParser.safeParse(jsonError);
      if (result2.success) {
        return Response.json(
          { message: result2.data.message },
          { status: response.status }
        );
      }
      console.error(jsonError);
      return Response.json(
        {
          message: `Frame server returned an unexpected error.`
        },
        { status: 500 }
      );
    }
    if (response.status === 302) {
      const location = response.headers.get("location");
      if (!location) {
        return Response.json(
          {
            message: "Frame server returned a redirect without a location header"
          },
          { status: 500 }
        );
      }
      return Response.json(
        {
          location
        },
        { status: 302 }
      );
    } else if (isPostRedirect) {
      return Response.json(
        {
          message: "Frame server did not return a 302 redirect"
        },
        { status: 500 }
      );
    }
    if (response.status >= 400 && response.status < 500) {
      const jsonError = await tryCallAsync(
        () => response.clone().json()
      );
      if (jsonError instanceof Error) {
        return Response.json(
          { message: jsonError.message },
          { status: response.status }
        );
      }
      const result2 = errorResponseParser.safeParse(jsonError);
      if (result2.success) {
        return Response.json(
          { message: result2.data.message },
          { status: response.status }
        );
      }
      console.error(jsonError);
      return Response.json(
        {
          message: `Frame server returned an unexpected error.`
        },
        { status: response.status }
      );
    }
    if (response.status !== 200) {
      return Response.json(
        {
          message: `Frame server returned a non-200 status code: ${response.status}`
        },
        { status: 500 }
      );
    }
    if (isTransactionRequest) {
      const transaction = await tryCallAsync(
        () => response.clone().json()
      );
      if (transaction instanceof Error) {
        return Response.json(
          { message: transaction.message },
          { status: 500 }
        );
      }
      return Response.json(transaction);
    }
    if (response.headers.get("content-type")?.toLowerCase().includes("application/json")) {
      const parseResult = await z.promise(jsonResponseParser).safeParseAsync(response.clone().json());
      if (!parseResult.success) {
        throw new Error("Invalid frame response");
      }
      const headers = new Headers(response.headers);
      headers.delete("content-encoding");
      return new Response(response.body, {
        headers,
        status: response.status,
        statusText: response.statusText
      });
    }
    const html = await response.text();
    if (multiSpecificationEnabled) {
      const result2 = await parseFramesWithReports({
        html,
        frameUrl: body.untrustedData.url,
        fallbackPostUrl: body.untrustedData.url,
        fromRequestMethod: "POST"
      });
      return Response.json(result2);
    }
    if (!isSpecificationValid(specification)) {
      return Response.json(
        {
          message: "Invalid specification"
        },
        { status: 400 }
      );
    }
    const result = await getFrame({
      htmlString: html,
      frameUrl: body.untrustedData.url,
      url: body.untrustedData.url,
      fromRequestMethod: "POST",
      specification
    });
    return Response.json(result);
  } catch (err) {
    console.error(err);
    return Response.json({ message: String(err) }, {
      status: 500
    });
  }
}

export {
  POST
};
