import { UseFrameAppReturn, UseFrameAppOptions } from '../use-frame-app.cjs';
import { HostEndpointEmitter } from './types.cjs';
import 'frames.js/frame-parsers';
import '@farcaster/frame-host';
import 'ox/Provider';
import 'ox/RpcSchema';

type UseFrameAppReturnSuccess = Extract<UseFrameAppReturn, {
    status: "success";
}>;
type UseFrameAppInIframeReturn = Exclude<UseFrameAppReturn, {
    status: "success";
}> | ({
    iframeProps: {
        src: string | undefined;
        ref: React.MutableRefObject<HTMLIFrameElement | null>;
    };
    emitter: HostEndpointEmitter;
} & Omit<UseFrameAppReturnSuccess, "emitter">);
/**
 * Handles frame app in iframe.
 *
 * On unmount it automatically unregisters the endpoint listener.
 *
 * @example
 * ```
 * import { useFrameAppInIframe } from '@frames.js/render/frame-app/iframe';
 * import { useWagmiProvider } from '@frames.js/render/frame-app/provider/wagmi';
 * import { useFarcasterSigner } from '@frames.js/render/identity/farcaster';
 *
 * function MyAppDialog() {
 *  const provider = useWagmiProvider();
 *  const farcasterSigner = useFarcasterSigner({
 *    // ...
 *  });
 *  const frameApp = useFrameAppInIframe({
 *    provider,
 *    farcasterSigner,
 *    // frame returned by useFrame() hook
 *    frame: frameState.frame,
 *    // ... handlers for frame app actions
 *  });
 *
 *  if (frameApp.status !== 'success') {
 *   // render loading or error
 *   return null;
 *  }
 *
 *  return <iframe {...frameApp.iframeProps} />;
 * }
 * ```
 */
declare function useFrameAppInIframe(options: UseFrameAppOptions): UseFrameAppInIframeReturn;

export { type UseFrameAppInIframeReturn, useFrameAppInIframe };
