import { Context, SetPrimaryButtonOptions, AddFrame, HostEndpoint, SignIn, FrameHost } from '@farcaster/frame-host';
import { ParseFramesV2ResultWithFrameworkDetails } from 'frames.js/frame-parsers';
import { Provider } from 'ox/Provider';
import { ExtractRequest, Default } from 'ox/RpcSchema';

type FrameClientConfig = Context.ClientContext;
type SendTransactionRpcRequest = ExtractRequest<Default, "eth_sendTransaction">;
type SignMessageRpcRequest = ExtractRequest<Default, "personal_sign">;
type SignTypedDataRpcRequest = ExtractRequest<Default, "eth_signTypedData_v4">;
type EthProvider = Provider<undefined, Default> & {
    setEventHandlers: (handlers: SharedEthProviderEventHandlers) => void;
};
/**
 * Function called when the frame app requests sending transaction.
 *
 * If false is returned then the request is cancelled and user rejected error is thrown
 */
type OnSendTransactionRequestFunction = (request: SendTransactionRpcRequest) => Promise<boolean>;
/**
 * Function called when the frame app requests signing message.
 *
 * If false is returned signing is cancelled and user rejected error is thrown
 */
type OnSignMessageRequestFunction = (request: SignMessageRpcRequest) => Promise<boolean>;
/**
 * Function called when the frame app requests signing typed data.
 *
 * If false is returned then the request is cancelled and user rejected error is thrown
 */
type OnSignTypedDataRequestFunction = (request: SignTypedDataRpcRequest) => Promise<boolean>;
type SharedEthProviderEventHandlers = {
    onSendTransactionRequest: OnSendTransactionRequestFunction;
    onSignMessageRequest: OnSignMessageRequestFunction;
    onSignTypedDataRequest: OnSignTypedDataRequestFunction;
};
type FramePrimaryButton = SetPrimaryButtonOptions;
type OnPrimaryButtonSetFunction = (options: FramePrimaryButton, pressedCallback: () => void) => void;
/**
 * Returns false if user rejected the request, otherwise it returns the notification details
 */
type OnAddFrameRequestedFunction = (frame: ParseFramesV2ResultWithFrameworkDetails) => Promise<false | Required<AddFrame.AddFrameResult>>;
type OnEIP6963RequestProviderRequestedFunctionOptions = {
    endpoint: HostEndpoint;
};
/**
 * Function that must emit eip6963:announceProvider event on endpoint to announce available providers
 */
type OnEIP6963RequestProviderRequestedFunction = (options: OnEIP6963RequestProviderRequestedFunctionOptions) => unknown;
type OnSignInFunctionOptions = SignIn.SignInOptions & {
    frame: ParseFramesV2ResultWithFrameworkDetails;
};
type OnSignInFunction = (options: OnSignInFunctionOptions) => Promise<SignIn.SignInResult>;
type OnViewProfileFunction = FrameHost["viewProfile"];
/**
 * Function called when the frame app is being loaded and we need to resolve the client that renders the frame app
 */
type ResolveClientFunction = (options: {
    signal: AbortSignal;
}) => Promise<FrameClientConfig>;
type HostEndpointEmitter = Pick<HostEndpoint, "emit" | "emitEthProvider">;

export type { EthProvider, FrameClientConfig, FramePrimaryButton, HostEndpointEmitter, OnAddFrameRequestedFunction, OnEIP6963RequestProviderRequestedFunction, OnEIP6963RequestProviderRequestedFunctionOptions, OnPrimaryButtonSetFunction, OnSendTransactionRequestFunction, OnSignInFunction, OnSignInFunctionOptions, OnSignMessageRequestFunction, OnSignTypedDataRequestFunction, OnViewProfileFunction, ResolveClientFunction, SendTransactionRpcRequest, SharedEthProviderEventHandlers, SignMessageRpcRequest, SignTypedDataRpcRequest };
