import { WebViewProps, WebView } from 'react-native-webview';
import { LegacyRef } from 'react';
import { UseFrameAppReturn, UseFrameAppOptions } from '../use-frame-app.cjs';
import { HostEndpointEmitter } from './types.cjs';
import 'frames.js/frame-parsers';
import '@farcaster/frame-host';
import 'ox/Provider';
import 'ox/RpcSchema';

type UseFrameAppReturnSuccess = Extract<UseFrameAppReturn, {
    status: "success";
}>;
type UseFrameAppInWebViewReturn = Exclude<UseFrameAppReturn, {
    status: "success";
}> | (Omit<UseFrameAppReturnSuccess, "emitter"> & {
    emitter: HostEndpointEmitter;
    webViewProps: {
        source: WebViewProps["source"];
        onMessage: NonNullable<WebViewProps["onMessage"]>;
        ref: LegacyRef<WebView>;
    };
});
/**
 * useFrameApp() hook for react-native-webview handling
 *
 * On unmount it automatically unregisters the endpoint listener.
 *
 * @example
 * ```
 * import { useFrameAppInWebView } from '@frames.js/render/frame-app/web-view';
 * import { useWagmiProvider } from '@frames.js/render/frame-app/provider/wagmi';
 * import { useFarcasterSigner } from '@frames.js/render/identity/farcaster';
 *
 * function MyAppDialog() {
 *   const provider = useWagmiProvider();
 *   const farcasterSigner = useFarcasterSigner({
 *     // ...
 *   });
 *   const frameApp = useFrameAppInWebView({
 *    provider,
 *    farcasterSigner,
 *    // frame returned by useFrame() hook
 *    frame: frameState.frame,
 *    // ... handlers for frame app actions
 *  });
 *
 *  if (frameApp.status !== 'success') {
 *     // render loading or error
 *    return null;
 *  }
 *
 *  return <WebView {...frameApp.webViewProps} />;
 * }
 * ```
 */
declare function useFrameAppInWebView(options: UseFrameAppOptions): UseFrameAppInWebViewReturn;

export { type UseFrameAppInWebViewReturn, useFrameAppInWebView };
