import {
  useFreshRef
} from "./chunk-KTZM67PO.js";
import {
  CastActionUnexpectedResponseError
} from "./chunk-LOKN5BYP.js";
import {
  isCastActionFrameResponse,
  isCastActionMessageResponse,
  isParseFramesWithReportsResult,
  mergeSearchParamsToUrl,
  tryCall,
  tryCallAsync
} from "./chunk-RVPSX5WL.js";
import "./chunk-6C3VEZWH.js";

// src/use-cast-action.ts
import { useCallback, useEffect, useMemo, useReducer, useRef } from "react";
function useCastAction({
  castId,
  enabled = true,
  fetch: fetchFunction = defaultFetchFn,
  onError,
  proxyUrl,
  signer,
  postUrl,
  onFrameResponse,
  onMessageResponse,
  onInvalidResponse
}) {
  const onErrorRef = useFreshRef(onError);
  const fetchRef = useFreshRef(fetchFunction);
  const onFrameResponseRef = useFreshRef(onFrameResponse);
  const onMessageResponseRef = useFreshRef(onMessageResponse);
  const onInvalidResponseRef = useFreshRef(onInvalidResponse);
  const signerRef = useFreshRef(signer);
  const lastFetchActionArgRef = useRef(null);
  const [state, dispatch] = useReducer(castActionReducer, {
    status: "idle",
    enabled
  });
  const castIdRef = useFreshRef(castId);
  const fetchAction = useCallback(
    async (arg) => {
      const currentSigner = arg.signer;
      if (currentSigner?.status !== "approved" && currentSigner?.status !== "impersonating") {
        await signerRef.current.onSignerlessFramePress();
        return;
      }
      dispatch({ type: "loading-url" });
      const signedDataOrError = await tryCallAsync(
        () => signerRef.current.signCastAction(currentSigner.privateKey, {
          postUrl: arg.postUrl,
          fid: currentSigner.fid,
          castId: arg.castId
        })
      );
      if (signedDataOrError instanceof Error) {
        tryCall(() => onErrorRef.current?.(signedDataOrError));
        dispatch({ type: "error", error: signedDataOrError });
        return;
      }
      const proxiedUrl = mergeSearchParamsToUrl(
        arg.proxyUrl,
        new URLSearchParams({ postUrl: arg.postUrl })
      );
      const actionResponseOrError = await tryCallAsync(
        () => fetchRef.current(proxiedUrl, {
          method: "POST",
          headers: {
            "Content-Type": "application/json"
          },
          body: JSON.stringify(signedDataOrError),
          cache: "no-cache"
        })
      );
      if (actionResponseOrError instanceof Error) {
        tryCall(() => onErrorRef.current?.(actionResponseOrError));
        dispatch({ type: "error", error: actionResponseOrError });
        return;
      }
      if (!actionResponseOrError.ok) {
        const error2 = new Error(
          `Unexpected response status ${actionResponseOrError.status}`
        );
        tryCall(() => onErrorRef.current?.(error2));
        dispatch({ type: "error", error: error2 });
        return;
      }
      const actionResponseDataOrError = await tryCallAsync(
        () => actionResponseOrError.clone().json()
      );
      if (actionResponseDataOrError instanceof Error) {
        tryCall(() => onErrorRef.current?.(actionResponseDataOrError));
        dispatch({ type: "error", error: actionResponseDataOrError });
        return;
      }
      if (isCastActionFrameResponse(actionResponseDataOrError)) {
        tryCall(() => onFrameResponseRef.current?.(actionResponseDataOrError));
        const signedFrameDataOrError = await tryCallAsync(() => {
          return signerRef.current.signFrameAction({
            buttonIndex: 1,
            frameButton: {
              action: "post",
              label: "cast action",
              post_url: actionResponseDataOrError.frameUrl,
              target: actionResponseDataOrError.frameUrl
            },
            frameContext: {
              castId: castIdRef.current
            },
            signer: currentSigner,
            url: actionResponseDataOrError.frameUrl,
            type: "default"
          });
        });
        if (signedFrameDataOrError instanceof Error) {
          tryCall(() => onErrorRef.current?.(signedFrameDataOrError));
          dispatch({ type: "error", error: signedFrameDataOrError });
          return;
        }
        const proxiedFrameUrl = mergeSearchParamsToUrl(
          arg.proxyUrl,
          signedFrameDataOrError.searchParams,
          new URLSearchParams({
            multispecification: "true",
            postUrl: actionResponseDataOrError.frameUrl
          })
        );
        const frameResponseOrError = await tryCallAsync(() => {
          return fetchRef.current(proxiedFrameUrl, {
            method: "POST",
            headers: {
              "Content-Type": "application/json"
            },
            body: JSON.stringify(signedFrameDataOrError.body),
            cache: "no-cache"
          });
        });
        if (frameResponseOrError instanceof Error) {
          tryCall(() => onErrorRef.current?.(frameResponseOrError));
          dispatch({ type: "error", error: frameResponseOrError });
          return;
        }
        if (!frameResponseOrError.ok) {
          const error2 = new Error(
            `Unexpected response status ${frameResponseOrError.status}`
          );
          tryCall(() => onErrorRef.current?.(error2));
          dispatch({ type: "error", error: error2 });
          return;
        }
        const frameResponseDataOrError = await tryCallAsync(
          () => frameResponseOrError.clone().json()
        );
        if (frameResponseDataOrError instanceof Error) {
          tryCall(() => onErrorRef.current?.(frameResponseDataOrError));
          dispatch({ type: "error", error: frameResponseDataOrError });
          return;
        }
        if (!isParseFramesWithReportsResult(frameResponseDataOrError)) {
          const error2 = new Error("Invalid frame response");
          tryCall(() => onErrorRef.current?.(error2));
          dispatch({ type: "error", error: error2 });
          return;
        }
        dispatch({
          type: "frame",
          response: actionResponseDataOrError,
          frame: frameResponseDataOrError
        });
        return;
      }
      if (isCastActionMessageResponse(actionResponseDataOrError)) {
        tryCall(
          () => onMessageResponseRef.current?.(actionResponseDataOrError)
        );
        dispatch({
          type: "message",
          response: actionResponseDataOrError
        });
        return;
      }
      tryCall(() => onInvalidResponseRef.current?.(actionResponseDataOrError));
      const error = new CastActionUnexpectedResponseError();
      tryCall(() => onErrorRef.current?.(error));
      dispatch({ type: "error", error });
    },
    [
      castIdRef,
      fetchRef,
      onErrorRef,
      onFrameResponseRef,
      onInvalidResponseRef,
      onMessageResponseRef,
      signerRef
    ]
  );
  const stateRef = useFreshRef(state);
  const refetch = useCallback(() => {
    if (!stateRef.current.enabled || !lastFetchActionArgRef.current) {
      return Promise.resolve();
    }
    return fetchAction(lastFetchActionArgRef.current);
  }, [fetchAction, stateRef]);
  useEffect(() => {
    dispatch({ type: "enabled-change", enabled });
  }, [enabled]);
  useEffect(() => {
    if (!enabled || !postUrl) {
      return;
    }
    lastFetchActionArgRef.current = {
      signer: signer.signer,
      postUrl,
      proxyUrl,
      castId
    };
    fetchAction(lastFetchActionArgRef.current).catch((e) => {
      onErrorRef.current?.(e instanceof Error ? e : new Error(String(e)));
    });
  }, [
    castId,
    postUrl,
    proxyUrl,
    signer.signer,
    enabled,
    fetchAction,
    onErrorRef
  ]);
  return useMemo(() => {
    switch (state.status) {
      case "idle":
        return {
          status: "idle",
          data: void 0,
          error: void 0,
          refetch
        };
      case "loading":
        return {
          status: "loading",
          data: void 0,
          error: void 0,
          refetch
        };
      case "error":
        return {
          status: "error",
          data: void 0,
          error: state.error,
          refetch
        };
      default: {
        if (state.type === "frame") {
          return {
            status: "success",
            type: "frame",
            data: state.response,
            frame: state.frame,
            error: void 0,
            refetch
          };
        }
        return {
          status: "success",
          type: "message",
          data: state.response,
          error: void 0,
          refetch
        };
      }
    }
  }, [state, refetch]);
}
function defaultFetchFn(url, init) {
  return fetch(url, init);
}
function castActionReducer(state, action) {
  if (action.type === "enabled-change") {
    if (action.enabled) {
      return {
        ...state,
        enabled: true
      };
    }
    return {
      status: "idle",
      enabled: false
    };
  }
  if (!state.enabled) {
    return state;
  }
  switch (action.type) {
    case "frame":
      return {
        ...state,
        status: "success",
        type: "frame",
        response: action.response,
        frame: action.frame
      };
    case "message":
      return {
        ...state,
        status: "success",
        type: "message",
        response: action.response
      };
    case "loading-url":
      return {
        ...state,
        status: "loading"
      };
    case "error":
      return {
        ...state,
        status: "error",
        error: action.error
      };
    default:
      return state;
  }
}
export {
  useCastAction
};
