import { ParseFramesV2ResultWithFrameworkDetails } from 'frames.js/frame-parsers';
import { Context, FrameHost, HostEndpoint } from '@farcaster/frame-host';
import { EthProvider, FrameClientConfig, ResolveClientFunction, OnPrimaryButtonSetFunction, OnAddFrameRequestedFunction, OnSignInFunction, OnViewProfileFunction, OnEIP6963RequestProviderRequestedFunction, OnSendTransactionRequestFunction, OnSignMessageRequestFunction, OnSignTypedDataRequestFunction, HostEndpointEmitter } from './frame-app/types.js';
import 'ox/Provider';
import 'ox/RpcSchema';

type UseFrameAppOptions = {
    /**
     * @example
     * ```ts
     * import { useWagmiProvider } from '@frames.js/render/frame-app/provider/wagmi';
     *
     * function Component() {
     *  const provider = useWagmiProvider();
     *  const frameApp = useFrameApp({
     *    provider,
     *  });
     *
     *  //...
     * }
     * ```
     */
    provider: EthProvider;
    /**
     * Frame client that is rendering the app
     *
     * This is async function if you need to fetch the client configuration
     * like notification settings, etc.
     *
     * Value should be memoized otherwise it will cause unnecessary re-renders.
     */
    client: FrameClientConfig | ResolveClientFunction;
    /**
     * Information about the context from which the frame was launched.
     *
     * @defaultValue launcher context
     */
    location?: Context.LocationContext;
    /**
     * Information about the user who manipulates with the frame.
     *
     * Value should be memoized otherwise it will cause unnecessary re-renders.
     */
    user: Context.UserContext;
    /**
     * Either:
     *
     * - frame parse result obtained from useFrame() hook
     * - frame url as string or URL object
     *
     * If value changes it resets the internal state and refetches the frame if the new value is url.
     */
    source: ParseFramesV2ResultWithFrameworkDetails | string | URL;
    /**
     * Custom fetch compatible function used to make requests.
     *
     * @see https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API
     */
    fetchFn?: typeof fetch;
    /**
     * Frame proxyUrl used to fetch the frame parse result.
     */
    proxyUrl: string;
    /**
     * Called when app calls `ready` method.
     */
    onReady?: FrameHost["ready"];
    /**
     * Called when app calls `close` method.
     */
    onClose?: FrameHost["close"];
    /**
     * Called when app calls `openUrl` method.
     */
    onOpenUrl?: FrameHost["openUrl"];
    /**
     * Called when app calls `setPrimaryButton` method.
     */
    onPrimaryButtonSet?: OnPrimaryButtonSetFunction;
    /**
     * Called when app calls `addFrame` method.
     *
     * If the frame manifest is invalid it is automatically rejected.
     * If the method has been called during the session more than once it immediatelly rejects
     */
    onAddFrameRequested?: OnAddFrameRequestedFunction;
    /**
     * Called when app calls `signIn` method.
     */
    onSignIn?: OnSignInFunction;
    /**
     * Called when app calls `viewProfile` method.
     */
    onViewProfile?: OnViewProfileFunction;
    /**
     * Called when app calls `eip6963RequestProvider` method.
     *
     * It will announce the provider to the frame app once this function returns the info
     */
    onEIP6963RequestProviderRequested?: OnEIP6963RequestProviderRequestedFunction;
    /**
     * Enabled debugging
     *
     * @defaultValue false
     */
    debug?: boolean;
    /**
     * Cache of frame app requests (calls to `addFrame` method).
     */
    addFrameRequestsCache?: Set<string>;
    onSendTransactionRequest?: OnSendTransactionRequestFunction;
    onSignMessageRequest?: OnSignMessageRequestFunction;
    onSignTypedDataRequest?: OnSignTypedDataRequestFunction;
};
type UseFrameAppReturn = {
    frame: ParseFramesV2ResultWithFrameworkDetails;
    client: FrameClientConfig;
    status: "success";
    /**
     * Creates sdk that must be exposed to frame app endpoint
     */
    sdk: (endpoint: HostEndpoint) => Omit<FrameHost, "ethProviderRequestV2">;
    /**
     * Gets emitter object that can be used to emit events to the frame app.
     */
    getEmitter: (endpoint: HostEndpoint) => HostEndpointEmitter;
} | {
    status: "pending";
} | {
    error: Error;
    status: "error";
};
/**
 * This hook is used to handle frames v2 apps.
 */
declare function useFrameApp({ provider, client, location, user, source, fetchFn, proxyUrl, onClose, onOpenUrl, onPrimaryButtonSet, onReady, debug, addFrameRequestsCache, onAddFrameRequested, onSendTransactionRequest, onSignMessageRequest, onSignTypedDataRequest, onViewProfile, onEIP6963RequestProviderRequested, onSignIn, }: UseFrameAppOptions): UseFrameAppReturn;

export { type UseFrameAppOptions, type UseFrameAppReturn, useFrameApp };
